import time
import kubernetes
import urllib3.exceptions
from kubernetes import client, config, watch
from cowait.network import get_remote_url
from cowait.tasks import TaskDefinition
from cowait.utils import json_stream
from cowait.engine.const import LABEL_TASK_ID, LABEL_PARENT_ID
from cowait.engine.cluster import ClusterProvider
from cowait.engine.errors import ProviderError, TaskCreationError
from cowait.engine.routers import create_router
from .task import KubernetesTask
from .volumes import create_volumes
from .utils import create_env, create_ports
from .affinity import create_affinity
from .pod import pod_is_ready, extract_pod_taskdef
from .errors import PodConfigError, PodUnschedulableError, PodTerminatedError, ImagePullError

DEFAULT_NAMESPACE = 'default'
DEFAULT_SERVICE_ACCOUNT = 'default'


class KubernetesProvider(ClusterProvider):
    def __init__(self, args={}):
        super().__init__('kubernetes', args)

        try:
            # attempt to load incluster config if available
            config.load_incluster_config()
        except kubernetes.config.ConfigException:
            # load local config
            config.load_kube_config(context=self.args.get('context', None))

        self.core = client.CoreV1Api()
        self.ext = client.ExtensionsV1beta1Api()
        self.custom = client.CustomObjectsApi()

        self.router = create_router(self, self.args.get('router', 'none'))

    @property
    def namespace(self):
        return self.args.get('namespace', DEFAULT_NAMESPACE)

    @property
    def service_account(self):
        return self.args.get('service_account', DEFAULT_SERVICE_ACCOUNT)

    @property
    def domain(self):
        return self.args.get('domain', None)

    @property
    def timeout(self):
        return self.args.get('timeout', 180)

    def spawn(self, taskdef: TaskDefinition) -> KubernetesTask:
        try:
            self.emit_sync('prepare', taskdef=taskdef)

            volumes, mounts = create_volumes(taskdef.volumes)

            # container definition
            container = client.V1Container(
                name=taskdef.id,
                image=taskdef.image,
                env=create_env(self, taskdef),
                ports=create_ports(taskdef.ports),
                image_pull_policy='Always',  # taskdef field??
                resources=client.V1ResourceRequirements(
                    requests={
                        'cpu': str(taskdef.cpu or '0'),
                        'memory': str(taskdef.memory or '0'),
                    },
                    limits={
                        'cpu': str(taskdef.cpu_limit or '0'),
                        'memory': str(taskdef.memory_limit or '0'),
                    },
                ),
                volume_mounts=mounts,
            )

            pod = self.core.create_namespaced_pod(
                namespace=self.namespace,
                body=client.V1Pod(
                    metadata=client.V1ObjectMeta(
                        name=taskdef.id,
                        namespace=self.namespace,
                        labels={
                            LABEL_TASK_ID: taskdef.id,
                            LABEL_PARENT_ID: taskdef.parent,
                            **taskdef.meta,
                        },
                    ),
                    spec=client.V1PodSpec(
                        hostname=taskdef.id,
                        restart_policy='Never',
                        image_pull_secrets=self.get_pull_secrets(),
                        volumes=volumes,

                        containers=[container],
                        service_account_name=self.service_account,
                        affinity=create_affinity(taskdef.affinity),
                    ),
                ),
            )

            # wrap & return task
            # print('~~ created kubenetes pod', pod.metadata.name)
            task = KubernetesTask(self, taskdef, pod)
            self.emit_sync('spawn', task=task)
            return task

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def kill(self, task_id):
        try:
            self.core.delete_collection_namespaced_pod(
                namespace=self.namespace,
                label_selector=f'{LABEL_TASK_ID}={task_id}',
            )
            self.emit_sync('kill', task_id=task_id)
            return task_id

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def get_task_pod(self, task_id):
        try:
            res = self.core.list_namespaced_pod(
                namespace=self.namespace,
                label_selector=f'{LABEL_TASK_ID}={task_id}',
            )
            return res.items[0] if len(res.items) > 0 else None

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def get_task_child_pods(self, task_id: str):
        try:
            res = self.core.list_namespaced_pod(
                namespace=self.namespace,
                label_selector=f'{LABEL_PARENT_ID}={task_id}',
            )
            return res.items

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def wait_until_ready(self, task_id: str, poll_interval: float = 1):
        while True:
            time.sleep(poll_interval)
            pod = self.get_task_pod(task_id)
            if not pod:
                raise ProviderError(f'No such task: {task_id}')

            try:
                if pod_is_ready(pod):
                    break
                else:
                    poll_interval = 1

            except PodUnschedulableError as e:
                poll_interval = 10
                print('warning: task', task_id, 'is unschedulable:', str(e))

            except PodTerminatedError:
                raise TaskCreationError('Task terminated') from None

            except ImagePullError:
                self.kill(task_id)
                raise TaskCreationError('Image pull failed') from None

            except PodConfigError:
                # todo: check pod events to figure out what went wrong, and report it back.
                # for now, leave the pod running so the user may inspect it
                raise TaskCreationError('Pod configuration error') from None

    def logs(self, task_id: str):
        # wait for pod to become ready
        self.wait_until_ready(task_id)

        try:
            w = watch.Watch()
            logs = w.stream(
                self.core.read_namespaced_pod_log,
                name=task_id,
                namespace=self.namespace,
            )

            def add_newline_stream():
                for log in logs:
                    yield log + '\n'

            return json_stream(add_newline_stream())

        except Exception:
            return self.logs(task_id)

    def destroy_all(self) -> list:
        try:
            self.core.delete_collection_namespaced_pod(
                namespace=self.namespace,
                label_selector=LABEL_TASK_ID,
            )

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def list_all(self) -> list:
        try:
            res = self.core.list_namespaced_pod(
                namespace=self.namespace,
                label_selector=LABEL_TASK_ID,
            )
            running = filter(lambda pod: pod.status.phase == 'Running', res.items)
            return [
                KubernetesTask(self, extract_pod_taskdef(pod), pod)
                for pod in running
            ]

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def destroy(self, task_id) -> list:
        def kill_family(task_id):
            # print('~~ kubernetes kill', task_id)

            kills = []
            children = self.get_task_child_pods(task_id)
            for child in children:
                child_id = child.metadata.labels[LABEL_TASK_ID]
                kills += kill_family(child_id)

            self.kill(task_id)
            kills.append(task_id)
            return kills

        return kill_family(task_id)

    def destroy_children(self, parent_id: str) -> list:
        try:
            # get a list of child pods
            children = self.core.list_namespaced_pod(
                namespace=self.namespace,
                label_selector=f'{LABEL_PARENT_ID}={parent_id}',
            )

            # destroy child pods
            self.core.delete_collection_namespaced_pod(
                namespace=self.namespace,
                label_selector=f'{LABEL_PARENT_ID}={parent_id}',
            )

            # return killed child ids
            return [
                child.metadata.labels[LABEL_TASK_ID]
                for child in children.items
            ]

        except urllib3.exceptions.MaxRetryError:
            raise ProviderError('Kubernetes engine unavailable')

    def get_pull_secrets(self):
        secrets = self.args.get('pull_secrets', [])
        return [client.V1LocalObjectReference(name=s) for s in secrets]

    def find_agent(self):
        pod = self.get_task_pod('agent')
        if pod is None:
            return None
        if pod.status.phase != 'Running':
            return None

        token = pod.metadata.labels['http_token']
        return get_remote_url(pod.status.pod_ip, token)

    def wait(self, task: KubernetesTask) -> bool:
        while True:
            pod = self.get_task_pod(task.id)
            if pod is None:
                return False

            if pod.status.phase == 'Running':
                time.sleep(2)
            elif pod.status.phase == 'Pending':
                time.sleep(2)
            elif pod.status.phase == 'Succeeded':
                return True
            else:
                return False



from kubernetes import client
from .router import Router
from ..const import LABEL_TASK_ID


class TraefikRouter(Router):
    def __init__(self, cluster):
        super().__init__(cluster)
        cluster.on('prepare', self.on_prepare)
        cluster.on('spawn', self.on_spawn)
        cluster.on('kill', self.on_kill)

    def on_prepare(self, taskdef):
        domain = self.cluster.domain
        if domain is None:
            raise RuntimeError('No cluster domain configured')

        for path, port in taskdef.routes.items():
            taskdef.routes[path] = {
                'port': port,
                'path': path,
                'url': f'http://{taskdef.id}.{domain}{path}',
            }
        return taskdef

    def on_spawn(self, task):
        ports = []
        rules = []

        idx = 0
        for path, route in task.routes.items():
            port = route['port']
            idx += 1
            port_name = f'route{idx}'

            ports.append(client.V1ServicePort(
                name=port_name,
                port=port,
                target_port=port,
            ))

            rules.append(client.ExtensionsV1beta1IngressRule(
                host=f'{task.id}.{self.cluster.domain}',
                http=client.ExtensionsV1beta1HTTPIngressRuleValue(
                    paths=[
                        client.ExtensionsV1beta1HTTPIngressPath(
                            path=path,
                            backend=client.ExtensionsV1beta1IngressBackend(
                                service_name=task.id,
                                service_port=port_name,
                            ),
                        ),
                    ],
                ),
            ))

        if len(rules) == 0:
            return

        print('~~ creating task ingress', path, '-> port', port)

        self.cluster.core.create_namespaced_service(
            namespace=self.cluster.namespace,
            body=client.V1Service(
                metadata=client.V1ObjectMeta(
                    name=task.id,
                    namespace=self.cluster.namespace,
                    labels={
                        LABEL_TASK_ID: task.id,
                    },
                ),
                spec=client.V1ServiceSpec(
                    selector={
                        LABEL_TASK_ID: task.id,
                    },
                    ports=ports,
                ),
            ),
        )

        self.cluster.ext.create_namespaced_ingress(
            namespace=self.cluster.namespace,
            body=client.ExtensionsV1beta1Ingress(
                metadata=client.V1ObjectMeta(
                    name=task.id,
                    labels={
                        LABEL_TASK_ID: task.id,
                    },
                    annotations={
                        'kubernetes.io/ingress.class': 'traefik',
                        'traefik.frontend.rule.type': 'PathPrefix',
                    },
                ),
                spec=client.ExtensionsV1beta1IngressSpec(
                    rules=rules,
                ),
            ),
        )

    def on_kill(self, task_id):
        try:
            self.cluster.core.delete_namespaced_service(
                namespace=self.cluster.namespace,
                name=task_id,
            )
        except client.rest.ApiException:
            pass

        try:
            self.cluster.ext.delete_namespaced_ingress(
                namespace=self.cluster.namespace,
                name=task_id
            )
        except client.rest.ApiException:
            pass



import docker
import requests.exceptions
from cowait.network import get_remote_url
from cowait.tasks import TaskDefinition
from cowait.utils import json_stream
from cowait.engine.const import LABEL_TASK_ID, LABEL_PARENT_ID
from cowait.engine.cluster import ClusterProvider
from cowait.engine.errors import ProviderError
from cowait.engine.routers import create_router
from .task import DockerTask
from .volumes import create_volumes
from .utils import create_env, create_ports, extract_container_taskdef

DEFAULT_NETWORK = 'cowait'


class DockerProvider(ClusterProvider):
    def __init__(self, args={}):
        super().__init__('docker', args)
        self.docker = docker.from_env()
        self.router = create_router(self, self.args.get('router', 'local'))

    @property
    def network(self):
        return self.args.get('network', DEFAULT_NETWORK)

    def spawn(self, taskdef: TaskDefinition) -> DockerTask:
        try:
            self.ensure_network()

            self.emit_sync('prepare', taskdef=taskdef)

            cpu_period = 100000
            cpu_quota = float(taskdef.cpu_limit or 0) * cpu_period

            container = self.docker.containers.run(
                detach=True,
                image=taskdef.image,
                name=taskdef.id,
                hostname=taskdef.id,
                network=self.network,
                ports=create_ports(taskdef),
                environment=create_env(self, taskdef),
                mounts=create_volumes(taskdef.volumes),
                cpu_quota=int(cpu_quota),
                cpu_period=int(cpu_period),
                mem_reservation=str(taskdef.memory or 0),
                mem_limit=str(taskdef.memory_limit or 0),
                labels={
                    LABEL_TASK_ID: taskdef.id,
                    LABEL_PARENT_ID: taskdef.parent,
                    **taskdef.meta,
                },
            )

            # print('~~ created docker container with id',
            #   container.id[:12], 'for task', taskdef.id)

            task = DockerTask(self, taskdef, container)
            self.emit_sync('spawn', task=task)
            return task

        except docker.errors.APIError as e:
            raise ProviderError(e.explanation)

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def wait(self, task: DockerTask) -> bool:
        """ Wait for a task to finish. Returns True on clean exit """
        result = task.container.wait()
        return result['StatusCode'] == 0

    def list_all(self) -> list:
        """ Returns a list of task definitions for all running tasks """
        try:
            containers = self.docker.containers.list(
                filters={'label': LABEL_TASK_ID},
            )
            return [extract_container_taskdef(c) for c in containers]

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy_all(self) -> None:
        """ Destroys all running tasks """
        try:
            containers = self.docker.containers.list(
                all=True,
                filters={
                    'label': LABEL_TASK_ID,
                },
            )

            for container in containers:
                container.remove(force=True)

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def find_child_containers(self, parent_id: str) -> list:
        """ Finds all child containers of a given task id """
        try:
            return self.docker.containers.list(
                filters={
                    'label': f'{LABEL_PARENT_ID}={parent_id}',
                },
            )

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy_children(self, parent_id: str) -> list:
        """ Destroy all child tasks of a given task id """
        try:
            children = self.find_child_containers(parent_id)

            tasks = []
            for child in children:
                tasks += self.destroy(child.labels[LABEL_TASK_ID])

            return tasks

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def destroy(self, task_id):
        """ Destroy a specific task id and all its descendants """

        # optimization: grab a list of all tasks at once, instead of querying
        # for every child.

        def kill_family(container):
            container_task_id = container.labels[LABEL_TASK_ID]
            # print(f'~~ kill {container_task_id} ({container.short_id})')

            children = self.find_child_containers(container_task_id)
            kills = []
            for child in children:
                kills += kill_family(child)

            try:
                container.remove(force=True)
            except docker.errors.NotFound:
                pass
            except docker.errors.APIError as e:
                if 'already in progress' in str(e):
                    pass
                else:
                    raise e

            kills.append(task_id)
            return kills

        try:
            container = self.docker.containers.get(task_id)
            return kill_family(container)

        except docker.errors.NotFound:
            return [task_id]

        except requests.exceptions.ChunkedEncodingError:
            # workaround for a bug in docker on mac:
            # https://github.com/docker/docker-py/issues/2696
            return None

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def logs(self, task_id: str):
        """ Stream task logs """
        try:
            container = self.docker.containers.get(task_id)
            return json_stream(container.logs(stream=True))

        except docker.errors.NotFound:
            raise ProviderError(f'No such task: {task_id}')

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def ensure_network(self):
        try:
            self.docker.networks.get(self.network)
        except docker.errors.NotFound:
            print('~~ creating docker network', self.network)
            self.docker.networks.create(
                name=self.network,
                check_duplicate=False,
                driver='bridge',
                labels={
                    'cowait': '1',
                })

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')

    def create_ports(self, taskdef):
        return taskdef.ports

    def find_agent(self):
        try:
            container = self.docker.containers.get('agent')
            if container.status != 'running':
                return None

            token = container.labels['http_token']
            return get_remote_url('agent', token)

        except docker.errors.NotFound:
            return None

        except requests.exceptions.ChunkedEncodingError:
            # workaround for a bug in docker on mac:
            # https://github.com/docker/docker-py/issues/2696
            return None

        except requests.exceptions.ConnectionError:
            raise ProviderError('Docker engine unavailable')




import ast
import json
import cowait
from cowait import Task
from .code_builder import CodeBuilder


class NotebookRunner(Task):
    async def run(self, path: str, **inputs):
        if not path.endswith('.ipynb'):
            path += '.ipynb'
        cells = file_to_json(path)['cells']

        code = CodeBuilder()

        code.append(0, 'async def _notebook_runner():')

        code.appendBlock(4, cells_to_code(cells))

        global_scope = {'cowait': cowait, 'NotebookRunner': self.__class__}
        local_scope = {}
        exec(str(code), global_scope, local_scope)
        handle = local_scope['_notebook_runner']
        return await handle()


def cells_to_code(cells):
    code = CodeBuilder()
    for cell in cells:
        cell_type = cell['cell_type']

        if cell_type == 'code':
            source_rows = [row[:-1] if row[-1] == '\n' else row for row in cell['source']]

            if len(source_rows) > 0:
                code.appendBlock(0, code_from_source(source_rows))
                code.append(0, '')

    return code


def code_from_source(source_rows):
    code = CodeBuilder()

    _, first_line = strip_indentation(source_rows[0])
    if first_line.startswith('%%'):
        command, *args = first_line[2:].split(' ', 1)
        new_code = transform_cell_magic(source_rows[1:], command, args)
        if new_code:
            code.appendBlock(0, new_code)
    else:
        for row in source_rows:
            indentation, line = strip_indentation(row)
            if line.startswith('global '):
                new_line = 'nonlocal ' + line[7:]
                code.append(indentation, new_line)
                print(f"Warning: Replaced '{line}' with '{new_line}'")
            elif line.startswith('%'):  # Not supported: a = %ls
                command, *args = line[1:].split(' ', 1)
                new_code = transform_line_magic(command, args)
                if new_code:
                    code.appendBlock(indentation, new_code)
            else:
                code.append(indentation, line)

    # If there is a syntax error it will be found here
    try:
        ast.parse(str(code), filename='<notebook cell>')
    except SyntaxError as e:
        syntax_error = SyntaxError(f"{e.msg}\nThe error is located somewhere in this cell:\n\n{str(code)}", e.args[1])
        raise syntax_error from None

    return code


def strip_indentation(row):
    line = row.lstrip(' ')
    return len(row) - len(line), line


def transform_line_magic(command: str, args: str):
    ignorables = ['lsmagic', 'matplotlib']

    if command in ignorables:
        return None

    raise ValueError(f"Magic command %{command} is not supported")


def transform_cell_magic(rows: list, command: str, args: str):
    ignorables = ['html', 'HTML', 'markdown', 'latex']

    if command in ignorables:
        return None

    raise ValueError(f"Magic command %%{command} is not supported")


def file_to_json(path):
    with open(path, 'r') as f:
        return json.load(f)



import os
import json
from cowait.tasks import TaskDefinition
from cowait.engine import get_cluster_provider, env_unpack, \
    ENV_TASK_CLUSTER, ENV_TASK_DEFINITION, ENV_GZIP_ENABLED


def env_get(key):
    if key not in os.environ:
        raise ValueError(f'Missing required environment variable {key}')
    value = os.environ[key]
    if os.getenv(ENV_GZIP_ENABLED, '0') == '1':
        return env_unpack(value)
    return json.loads(value)


def env_get_cluster_provider():
    clusterdef = env_get(ENV_TASK_CLUSTER)
    return get_cluster_provider(**clusterdef)


def env_get_task_definition():
    taskdef = env_get(ENV_TASK_DEFINITION)
    return TaskDefinition.deserialize(taskdef)



import json


class StoppedError(RuntimeError):
    def __init__(self, result):
        super().__init__('Exited')
        self.result = result

    def _render_traceback_(self):
        try:
            return ['Returned with result:'] + json.dumps(self.result, indent=2).split('\n')
        except TypeError:
            return ['Returned with result:', str(self.result)]


class TaskError(RuntimeError):
    """
    Raised when an error is received from a subtask

    Attributes:
        error (str): Original error message
    """

    def __init__(self, error: str):
        """
        Arguments:
            error (str): Original error message
        """
        self.error = error


class TaskNotFoundError(RuntimeError):
    """ Raised when a task can not be instantiated """
    pass



from datetime import datetime
import json
from sty import fg, rs
from time import time as ctime
from .utils import printheader


class Logger(object):
    def __init__(self, quiet: bool = False, time: bool = True):
        self.quiet = quiet
        self.time = time
        self.start = ctime()

    def header(self, title: str = None):
        if self.quiet:
            return
        printheader(title)

    @property
    def newline_indent(self):
        return 0 if not self.time else 6

    def print(self, *args):
        if self.quiet:
            return
        text = ' '.join(map(lambda s: str(s), args))
        last_break = text[-1] == '\n'
        if last_break:
            text = text[:-1]
        if self.newline_indent > 0:
            text = text.replace('\n', '\n'.ljust(self.newline_indent))
        print(text, end='' if not last_break else '\n')

    def println(self, *args, time=False):
        text = ' '.join(map(lambda s: str(s), args))
        if len(text) == 0 or text[-1] != '\n':
            text += '\n'
        self.print(text)

    def print_time(self, ts: str = None):
        if not self.time:
            return
        if ts is None:
            elapsed = ctime() - self.start
            self.print(f'{elapsed:05.1f} ')
        else:
            dt = datetime.fromisoformat(ts)
            timestamp = dt.strftime('%H:%M:%S')
            self.print(f'{timestamp} ')

    def print_exception(self, error):
        self.header(f'error')
        self.println(error)

    def json(self, obj, indent: int = 0, lv=0):
        idt = indent * lv * ' '
        keyc = fg.blue
        strc = fg.green
        numc = fg.red
        dotc = rs.all

        if isinstance(obj, dict):
            if len(obj) > 2 and len(obj) < 20 and indent > 0:
                return f'{idt}{dotc}{{{rs.all}\n' + \
                    f'\n{idt}'.join([
                        f'{keyc}{k}{dotc}:{rs.all} '+self.json(v, lv=lv+1, indent=indent)
                        for k, v in obj.items()
                    ]) + \
                    f'\n{idt}{keyc}}}{rs.all}'
            return f'{idt}{dotc}{{{rs.all} ' + \
                ', '.join([
                    f'{keyc}{k}{dotc}:{rs.all} '+self.json(v, lv=lv+1, indent=indent)
                    for k, v in obj.items()
                ]) + f' {dotc}}}{rs.all}'
        elif isinstance(obj, list):
            if len(obj) > 2 and len(obj) < 20 and indent > 0:
                return f'{idt}{dotc}[{rs.all}\n' + \
                    f'{dotc},{rs.all}\n{idt}'.join([
                        self.json(v, lv=lv+1, indent=indent)
                        for v in obj
                    ]) + f'\n{idt}{dotc}]{rs.all}'
            return f'{dotc}[{rs.all} ' + \
                f'{dotc},{rs.all} '.join([
                    self.json(v, lv=lv+1, indent=indent)
                    for v in obj
                ]) + f' {dotc}]{rs.all}'
        elif isinstance(obj, int):
            return f'{numc}{obj}{rs.all}'
        elif isinstance(obj, str):
            return f'{strc}' + json.dumps(obj) + rs.all
        elif isinstance(obj, bool):
            return f'{numc}true{rs.all}' if obj else f'{numc}false{rs.all}'
        else:
            return json.dumps(obj)



import json
from datetime import datetime
from .logger import Logger


class JSONLogger(Logger):
    def log(self, type: str, msg: dict) -> None:
        msg['ts'] = datetime.now().isoformat()
        self.print(json.dumps(msg), flush=True)



import traceback
from datetime import datetime
from cowait.tasks import Task, TaskError, rpc, sleep
from cowait.utils import parse_task_image_name
from .schedule_definition import ScheduleDefinition


def scheduled_task_id(name: str, date: datetime):
    name = name.replace('.', '-')
    name = name.replace('_', '-')
    name = name.lower()
    timestamp = date.strftime('%Y%m%d%H%M')
    return f'{name}-{timestamp}'


class ScheduleTask(Task):
    """
    ScheduleTask runs another task on a specified cron-style schedule.

    Inputs:
        schedule (str): Cron-style scheduling string, e.g. '0 8 * * *' for every day at 8am.
        target (str): Task to run.
        kwargs: All other inputs will be forwarded to the target task.

    Returns:
        Nothing. Runs forever.
    """

    async def run(self, schedule: str, target: str, **inputs):
        image, name = parse_task_image_name(target, default_image=self.image)

        # instantiate schedule
        self.schedule = ScheduleDefinition(schedule)

        print(f'Scheduled {name} ({image}) at {schedule}')
        print('The current time is', datetime.now())

        last_run = datetime(1970, 1, 1)
        while True:
            await sleep(1)

            now = datetime.now()
            if self.schedule.is_now():
                if self.schedule.is_at(last_run) \
                        and last_run.minute == now.minute:
                    continue

                last_run = now
                await self.execute(name, image, inputs)

    async def execute(self, name: str, image: str, inputs: dict):
        now = datetime.now()
        print('Execution triggered')

        try:
            result = await self.spawn(
                id=scheduled_task_id(name, now),
                name=name,
                image=image,
                inputs=inputs,
            )

            end = datetime.now()
            duration = end - now
            print(f'Finished in {duration}. Result:')
            print(result)

        except TaskError:
            end = datetime.now()
            duration = end - now
            print(f'Error after {duration}. Exception:')
            traceback.print_exc()

    @rpc
    async def reschedule(self, schedule):
        # instantiate/validate schedule
        self.schedule = ScheduleDefinition(schedule)
        return schedule



import json


def json_stream(stream):
    buffer = ''
    for log in stream:
        if isinstance(log, str):
            buffer += log
        elif isinstance(log, bytes):
            buffer += str(log, encoding='utf-8')
        else:
            raise TypeError('Cant append ' + type(log) + ' to buffer')

        while '\n' in buffer:
            split = buffer.find('\n')
            chunk = buffer[:split]
            buffer = buffer[split+1:]

            if len(chunk.strip()) == 0:
                continue

            try:
                yield json.loads(chunk)
            except json.JSONDecodeError:
                yield {
                    'type': 'core/error',
                    'error': 'json decode error',
                    'chunk': chunk,
                }



import os
import json
from cowait.tasks import Task, sleep


class UtilityTask(Task):
    """
    Test utility container
    """

    def __init__(self, taskdef, cluster, node):
        super().__init__(
            taskdef=taskdef,
            cluster=cluster,
            node=node,
        )
        # store taskdef for later comparison
        self.taskdef = taskdef

    async def run(
        self,
        child: bool = False,
        error: bool = False,
        child_error: bool = False,
        **inputs
    ):
        # dump task information to stdout
        print(json.dumps({
            'taskdef': self.taskdef.serialize(),
            'env': dict(os.environ),
        }))

        # create a child
        if child or child_error:
            print('spawn child')
            await UtilityTask(error=child_error)

        if error:
            raise RuntimeError('Caused test error')

        # run forever
        if inputs.get('forever', False):
            while True:
                await sleep(1)

        # return desired output
        return inputs.get('return', {})



import os  # stinky, dont read envs here
import asyncio
from concurrent.futures import Future
from pyspark.conf import SparkConf
from cowait.tasks import Task, sleep, rpc
from cowait.tasks.messages import TASK_LOG
from .master import SparkMaster
from .worker import SparkWorker


MSG_LEADER = 'I have been elected leader!'
MSG_REGISTER = 'Successfully registered with master'

SPARK_PACKAGES = [
    'com.amazonaws:aws-java-sdk-pom:1.11.375',
    'org.apache.hadoop:hadoop-aws:3.2.0',
]

SPARK_DEFAULTS = {
    'spark.executor.instances': '2',
    'spark.dynamicAllocation.enabled': 'true',

    # packages
    'spark.jars.packages': ','.join(SPARK_PACKAGES),

    # hadoop config
    'spark.hadoop.fs.s3a.access.key': os.getenv('AWS_ACCESS_KEY_ID'),
    'spark.hadoop.fs.s3a.secret.key': os.getenv('AWS_SECRET_ACCESS_KEY'),
    'spark.hadoop.fs.s3a.endpoint': 's3-eu-west-1.amazonaws.com',

    # 'spark.hadoop.fs.s3a.committer.magic.enabled': 'true',
    # 'spark.hadoop.fs.s3a.committer.name': 'magic',

    'spark.hadoop.fs.s3a.path.style.access': 'true',
    'spark.hadoop.fs.s3a.fast.upload': 'true',
    'spark.hadoop.fs.s3a.fast.upload.buffer': 'bytebuffer',
    'spark.hadoop.fs.s3a.path.style.access': 'true',
    'spark.hadoop.fs.s3a.multipart.size': '128M',
    'spark.hadoop.fs.s3a.fast.upload.active.blocks': '4',
    'spark.hadoop.fs.s3a.committer.name': 'partitioned',
}


class SparkCluster(Task):
    def init(self):
        self.master = None
        self.workers = []
        self.spark_cluster = {}
        self.running = True
        self.app_name = self.id

        # subscribe to logs
        self.node.server.on(TASK_LOG, self.on_log)

    async def on_log(self, id, file, data, **msg) -> None:
        if self.master and id == self.master.id:
            if MSG_LEADER in data:
                print('~~ spark master ready')
                self.master.ready.set_result(self.master)

        for worker in self.workers:
            if id == worker.id and MSG_REGISTER in data:
                print(f'~~ spark worker {worker.id} ready')
                worker.ready.set_result(worker)

        return False

    async def before(self, inputs: dict) -> dict:
        # create cluster
        await self.setup_cluster(
            workers=inputs.get('workers'),
        )
        return inputs

    async def run(self, app_name: str = None, workers: int = 2):
        if app_name is not None:
            self.app_name = app_name

        if '/' in self.master.routes:
            print('Spark WebUI available at:')
            print(self.master.routes['/']['url'])

        while self.running:
            await sleep(1)

        return {}

    async def after(self, inputs: dict):
        print('~~ destroying spark cluster')
        self.master.destroy()
        for worker in self.workers:
            worker.destroy()

        await super().after(inputs)

    async def setup_cluster(self, workers=2, **config) -> None:
        print('~~ creating spark cluster...')
        print(f'~~   workers = {workers}')

        # create spark master
        self.master = self.spawn(SparkMaster, routes={'/': 8080})
        self.master.ready = Future()
        self.master_uri = f'spark://{self.master.ip}:7077'

        await sleep(1)

        # create spark workers
        self.workers = []
        await self.add_workers(workers)

        print('~~ waiting for cluster nodes')
        await self.wait_for_nodes()

        print('~~ spark cluster ready')

    @rpc
    async def get_workers(self):
        ready = filter(lambda w: w.ready.done(), self.workers)
        ids = list(map(lambda w: w.id, ready))
        return {
            'workers': ids,
            'ready': len(ids),
            'total': len(self.workers),
        }

    @rpc
    async def scale(self, workers: int):
        diff = workers - len(self.workers)
        if diff > 0:
            # scale up
            print(f'~~ scale({workers}): adding {diff} workers')
            await self.add_workers(diff)

        elif diff < 0:
            # scale down
            print(f'~~ scale({workers}): removing {abs(diff)} workers')
            await self.remove_workers(diff)

    @rpc
    async def get_config(self) -> SparkConf:
        conf = SparkConf() \
            .setAppName(self.app_name) \
            .setMaster(self.master_uri)
        for option, value in SPARK_DEFAULTS.items():
            conf.set(option, value)
        return conf

    async def add_workers(self, count):
        for i in range(0, count):
            w = self.spawn(
                SparkWorker,
                routes={},
                ports={},
                master=self.master_uri,
                cores=2,
            )
            w.ready = Future()
            self.workers.append(w)

    async def remove_workers(self, count):
        count = abs(count)
        if count > len(self.workers):
            count = len(self.workers)

        for worker in self.workers[-count:]:
            worker.destroy()
        self.workers = self.workers[:-count]

    async def wait_for_nodes(self):
        await asyncio.gather(
            asyncio.wrap_future(self.master.ready),
            *map(lambda w: asyncio.wrap_future(w.ready), self.workers),
        )

    @rpc
    async def teardown(self) -> None:
        self.running = False



import os
import asyncio
from cowait.tasks import Task, rpc
from cowait.types import Dict

ShellResult = Dict({
    'code': int,
})


class ShellTask(Task):
    async def run(self, command: str, env: dict = {}) -> ShellResult:
        # run shell command
        self.process = await asyncio.create_subprocess_shell(
            command,
            stderr=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            env={
                **os.environ,
                **env,
            }
        )

        # setup stream readers
        print_stream(self.process.stdout, 'stdout', self.filter_stdout)
        print_stream(self.process.stderr, 'stderr', self.filter_stderr)

        # wait for process to finish
        result = await self.process.wait()

        return {
            'code': result,
        }

    def filter_stdout(self, line: str) -> bool:
        return True

    def filter_stderr(self, line: str) -> bool:
        return True

    @rpc
    async def stop(self):
        print('Stopping shell process')
        self.process.kill()
        return await super().stop()


def print_stream(stream, name: str, filter: callable) -> None:
    async def logger() -> None:
        while True:
            line = await stream.readline()
            if line == b'':
                return

            text = line.decode('utf-8')
            if filter is not None:
                if not filter(text):
                    continue

            print(text, end='')

    asyncio.create_task(logger())



import click
import cowait.cli.commands
from .utils import parse_input_list


@click.group(help='cluster management')
@click.pass_context
def cluster(ctx):
    pass


@cluster.command(help='describe cluster')
@click.argument('name', type=str)
@click.pass_context
def get(ctx, name: str):
    cowait.cli.cluster_get(ctx.obj, name)


@cluster.command(help='list all clusters')
@click.pass_context
def ls(ctx):
    cowait.cli.cluster_ls(ctx.obj)


@cluster.command(help='default cluster name')
@click.pass_context
def default(ctx):
    cowait.cli.cluster_default(ctx.obj)


@cluster.command(help='set default cluster')
@click.argument('name', type=str)
@click.pass_context
def set_default(ctx, name: str):
    cowait.cli.cluster_set_default(ctx.obj, name)


@cluster.command(help='add new cluster')
@click.argument('name', type=str)
@click.option('--type', type=str, help='cluster provider type')
@click.option('-o', '--option',
              type=str,
              multiple=True,
              help='specify cluster provider option')
@click.pass_context
def add(ctx, name: str, type: str, option: dict = {}):
    cowait.cli.cluster_add(ctx.obj, name, type.lower(), **parse_input_list(option))


@cluster.command(help='remove cluster')
@click.argument('name', type=str)
@click.pass_context
def rm(ctx, name: str):
    cowait.cli.cluster_rm(ctx.obj, name)



import inspect
from functools import reduce
from typing import ClassVar
from .type import Type


_type_mapping = {}
_name_mapping = {}


def register_type(from_type: ClassVar, to_type: Type) -> None:
    """ Registers a type alias for the given from_type """
    global _name_mapping
    global _type_mapping
    if to_type.name is None:
        to_type.name = f'{to_type.__module__}.{to_type.__name__}'
    _name_mapping[to_type.name] = to_type
    _type_mapping[from_type] = to_type
    _name_mapping[f'{to_type.__module__}.{to_type.__name__}'] = to_type


def get_type(name: str) -> Type:
    global _name_mapping
    if name in _name_mapping:
        return _name_mapping[name]
    raise TypeError(f'Unknown type {name}')


def TypeAlias(*aliases: ClassVar):
    """ Registers a type alias for the decorated type """
    def register_alias(cls: ClassVar):
        if not is_cowait_type(cls):
            raise TypeError(f'{cls} must be a valid cowait type')

        for alias in aliases:
            register_type(alias, cls)

        return cls

    return register_alias


def is_cowait_type(type: any) -> bool:
    """ Checks if a type is a valid cowait type """
    global _type_mapping

    if isinstance(type, Type):
        return True
    elif type in _type_mapping:
        # type can be considered a cowait type if we have
        # a registered mapping for it.
        return True
    elif inspect.isclass(type) and issubclass(type, Type):
        # classes that can be instantiated to type objects
        # are also considered valid.
        return True

    return False


def convert_type(type: any) -> Type:
    """ Converts a python type to a cowait type """
    global _type_mapping

    if isinstance(type, Type):
        # if passed a valid Cowait type instance, simply return it
        return type

    elif type in _type_mapping:
        # check if we have a registered mapping for this type.
        # if so, instantiate that the mapped type.
        mapped_type = _type_mapping[type]
        return instantiate_type(mapped_type)

    elif inspect.isclass(type) and issubclass(type, Type):
        # if passed a valid Cowait type class, attempt to instantiate it
        return instantiate_type(type)

    else:
        raise TypeError(f'Expected {type} to be a valid cowait type')


def param_without_default(p: inspect.Parameter) -> bool:
    """ Checks if a function parameter has no default value """

    # self parameter does not need a default.
    if p.name == 'self':
        return False

    # variable arguments do not need defaults
    if p.kind == inspect._VAR_KEYWORD or \
       p.kind == inspect._VAR_POSITIONAL:
        return False

    return p.default == inspect.Parameter.empty


def instantiate_type(TypeClass: ClassVar) -> object:
    """ Attempts to automatically instantiate a cowait type class """

    # count number of __init__ parameters without a default value
    sig = inspect.signature(TypeClass.__init__)

    non_default_params = reduce(
        lambda c, p: c+1 if param_without_default(p) else c,
        sig.parameters.values(),
        0,
    )

    # can't auto instantiate if there are non-default parameters
    if (non_default_params > 0):
        raise TypeError(f'{TypeClass.__name__} has non-default parameters '
                        f'and cant be automatically instantiated')

    instance = TypeClass()
    return instance





class CodeBuilder:
    def __init__(self):
        self.lines = []
    
    def append(self, indentation: int, line: str):
        if len(line) > 0 and line[-1] != '\n':
            self.lines.append(' ' * indentation + line)
        else:
            self.lines.append(' ' * indentation + line[:-1])
        return self

    def appendBlock(self, indentation: int, block):
        for line in block.lines:
            self.append(indentation, line)
        return self
    
    def __str__(self):
        s = ''
        for line in self.lines:
            s += line
            s += '\n'
        return s



import os
import sys
import time
import getpass
import platform
import traceback
import subprocess
from kubernetes import client
from cowait.cli.utils import ExitTrap
from cowait.cli.context import Context
from cowait.cli.commands.run import RunLogger, run as run_cmd
from cowait.tasks import TaskDefinition
from cowait.utils import uuid


def notebook(config, image: str = None, cluster_name: str = None) -> None:
    context = Context.open(config)

    if not context.notebook:
        print('Notebook funcitonaility is not enabled.')
        print('To enable, set features.notebook to True in cowait.yml and rebuild.')
        sys.exit(1)

    if image is not None:
        print('Remote images are currently not supported')
        sys.exit(1)

    volumes = {
        '/var/task': {
            'bind': {
                'src': os.getcwd(),
                'mode': 'rw',
                'inherit': 'same-image',
            },
        }
    }

    cluster = context.get_cluster(cluster_name)

    # Docker
    if cluster.type == 'docker':
        return run_cmd(
            config=config,
            task='cowait.notebook',
            build=False,
            image=image,
            routes={
                '/': '8888',
            },
            cluster_name=cluster_name,
            volumes=volumes,
        )

    # check for clientfs
    clientfs_executable = './clientfs-' + platform.system().lower()
    if not os.path.exists(clientfs_executable):
        print('Kubernetes notebooks are not supported in this build of Cowait')
        sys.exit(1)

    # Kubernetes
    core = client.CoreV1Api()
    notebook_id = 'notebook-' + uuid(4)

    core.create_namespaced_persistent_volume_claim(
        namespace=cluster.namespace,
        body=client.V1PersistentVolumeClaim(
            metadata=client.V1ObjectMeta(
                name=notebook_id,
                namespace=cluster.namespace,
            ),
            spec=client.V1PersistentVolumeClaimSpec(
                storage_class_name='clientfs',
                access_modes=['ReadWriteMany'],
                resources=client.V1ResourceRequirements(
                    requests={
                        'storage': '1G',
                    },
                ),
            ),
        ),
    )

    def delete_pvc(task_id):
        print('destroy', task_id)
        if task_id != notebook_id:
            return

        print('* stopping clientfs')
        clientfs.terminate()

        print('* deleting volume')
        core.delete_namespaced_persistent_volume_claim(notebook_id, cluster.namespace)

    cluster.on('kill', delete_pvc)

    pvc_id = None

    while True:
        time.sleep(1)
        volume = core.read_namespaced_persistent_volume_claim(notebook_id, cluster.namespace)
        if volume.status.phase == 'Bound':
            pvc_id = 'pvc-' + volume.metadata.uid
            print('* created volume', notebook_id, '/', pvc_id)
            break

    volumes['/var/task'] = {
        'persistent_volume_claim': {
            'claim_name': notebook_id,
        },
    }

    # start clientfs
    clientfs_host = cluster.args.get('clientfs', {}).get('host')
    print(f'* connecting clientfs volume to {clientfs_host}...')
    clientfs = subprocess.Popen([
        clientfs_executable,
        f"--proxy={clientfs_host}",
        f"--volume={pvc_id}"
    ])

    logger = RunLogger()
    try:
        # default to agent as upstream
        agent = cluster.find_agent()

        # create task definition
        taskdef = TaskDefinition(
            id=notebook_id,
            name='cowait.notebook',
            image=context.image,
            env={
                **context.extend('environment', {}),
                **context.dotenv,
            },
            routes={
                '/': '8888',
            },
            parent=None,  # root task
            upstream=agent,
            owner=getpass.getuser(),
            volumes=context.extend('volumes', volumes),
        )

        # print execution info
        logger.print_info(taskdef, cluster)

        # submit task to cluster
        task = cluster.spawn(taskdef)

        detach = False
        if detach:
            logger.header('detached')
            return

        def destroy(*args):
            logger.header('interrupt')
            cluster.destroy(task.id)
            sys.exit(1)

        with ExitTrap(destroy):
            # capture & print logs
            logs = cluster.logs(task.id)
            logger.header('task output')
            for msg in logs:
                logger.handle(msg)

    except Exception:
        traceback.print_exc()
        sys.exit(1)


def run_notebook(
    config, path: str,
    image: str, cluster: str, name: str, inputs: dict, env: dict,
    build: bool, detach: bool, quiet: bool
):
    run_cmd(
        config,
        task='cowait.tasks.notebook',
        name=name,
        image=image,
        inputs={
            **inputs,
            'path': path
        },
        env=env,
        ports={},
        routes={},
        build=build,
        detach=detach,
        quiet=quiet,
        cluster_name=cluster,
    )



from datetime import datetime


class ScheduleDefinition(object):
    def __init__(self, schedule):
        parts = schedule.split(' ')
        if len(parts) != 5:
            raise ValueError('Invalid schedule syntax')

        self.minute, self.hour, self.date, self.month, self.dow = parts

        # run a check to validate the schedule
        self.is_now()

    def is_now(self):
        return self.is_at(datetime.now())

    def is_at(self, time):
        if not schedule_match(self.minute, time.minute, 0, 59):
            return False
        if not schedule_match(self.hour, time.hour, 0, 23):
            return False
        if not schedule_match(self.date, time.day, 1, 31):
            return False
        if not schedule_match(self.month, time.month, 1, 12):
            return False
        if not schedule_match(self.dow, time.weekday(), 1, 7):
            return False
        return True


def schedule_match(pattern, value, min_val, max_val):
    def in_bounds(value):
        return value >= min_val and value <= max_val

    if pattern == '*':
        return True

    if ',' in pattern:
        patterns = pattern.split(',')
        for pattern in patterns:
            if schedule_match(pattern, value):
                return True
        return False

    if '-' in pattern:
        dash = pattern.find('-')
        minimum = int(pattern[:dash])
        if not in_bounds(minimum):
            raise ValueError(
                f'Range minimum {minimum} is out of range: '
                f'{min_val}-{max_val}')

        maximum = int(pattern[dash+1:])
        if not in_bounds(maximum):
            raise ValueError(
                f'Range maximum {maximum} is out of range: '
                f'{min_val}-{max_val}')

        return value >= minimum and value <= maximum

    if '/' in pattern:
        slash = pattern.find('/')
        setting = pattern[:slash]
        divisor = int(pattern[slash+1:])
        if setting == '*':
            return value % divisor == 0
        else:
            raise ValueError(
                "Illegal schedule divisor pattern. "
                "Only */n patterns are supported."
            )

    exact = int(pattern)
    if not in_bounds(exact):
        raise ValueError(
            f'Exact value {exact} is out of range: '
            f'{min_val}-{max_val}')

    return value == exact



from aiohttp import web
from cowait.utils import uuid
from .const import WS_PATH, API_PATH, RPC_PATH, QUERY_TOKEN


class AuthMiddleware(object):
    def __init__(self):
        self.tokens = {}
        self.enabled = True

    def add_token(self, token: str) -> None:
        self.tokens[token] = True

    def get_token(self) -> str:
        token = uuid(32)
        self.tokens[token] = True
        return token

    def ban_token(self, token: str) -> None:
        self.tokens[token] = False

    def validate_token(self, token: str) -> bool:
        return self.tokens.get(token, False)

    def is_public(self, path):
        if not self.enabled:
            return True

        if path == f'/{WS_PATH}':
            return False
        if path.startswith(f'/{API_PATH}/'):
            return False
        if path.startswith(f'/{RPC_PATH}/'):
            return False

        return True

    @web.middleware
    async def middleware(self, request, handler):
        # allow public routes
        if self.is_public(request.path):
            return await handler(request)

        # check for authorization header
        if 'authorization' in request.headers:
            # validate authorization header
            auth = request.headers.get('authorization')
            if 'bearer ' not in auth.lower():
                raise web.HTTPBadRequest(reason='Invalid authorization header')

            token = auth[len('bearer '):]
            if self.validate_token(token):
                return await handler(request)

        # check for query token
        if QUERY_TOKEN in request.query:
            token = request.query[QUERY_TOKEN]
            if self.validate_token(token):
                return await handler(request)

        raise web.HTTPUnauthorized()



import yaml
import click
import cowait.cli.commands
from cowait.cli import CliError
from .utils import parse_input_list


@click.command(help='run a task')
@click.argument('task', type=str)
@click.option('--image',
              default=None,
              type=str,
              help='image name')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-n', '--name',
              type=str,
              default=None,
              help='specific task name')
@click.option('-i', '--input',
              type=str,
              multiple=True,
              help='specify task input')
@click.option('-e', '--env',
              type=str,
              multiple=True,
              help='define enviornment variable')
@click.option('-p', '--port',
              type=int,
              multiple=True,
              help='open a port')
@click.option('-r', '--route',
              type=str,
              multiple=True,
              help='add an ingress route')
@click.option('-u', '--upstream',
              type=str,
              help='root task upstream uri')
@click.option('-b', '--build',
              type=bool, is_flag=True,
              help='build and push first',
              default=False)
@click.option('-d', '--detach',
              type=bool, is_flag=True,
              help='run in detached mode',
              default=False)
@click.option('--cpu',
              help='cpu request',
              type=str,
              default=None)
@click.option('--cpu-limit',
              help='cpu limit',
              type=str,
              default=None)
@click.option('--memory', '--mem',
              help='memory request',
              type=str,
              default=None)
@click.option('--memory-limit', '--mem-limit',
              help='memory limit',
              type=str,
              default=None)
@click.option('-f', '--json', '--yml', '--yaml', 'file',
              help='yaml/json file with inputs',
              type=str,
              default=None)
@click.option('--raw',
              type=bool, is_flag=True,
              help='raw json output',
              default=False)
@click.option('-q', '--quiet',
              type=bool, is_flag=True,
              help='no output except result',
              default=False)
@click.option('-a', '--affinity',
              help='task affinity (stack/spread)',
              type=str,
              default=None)
@click.pass_context
def run(
    ctx, task: str, image: str, cluster: str, name: str,
    input, env, port, route,
    upstream: str, build: bool, detach: bool,
    cpu: str, cpu_limit: str, memory: str, memory_limit: str,
    file: str, raw: bool, quiet: bool, affinity: str
):
    file_inputs = {}
    if file is not None:
        try:
            with open(file, 'r') as f:
                file_inputs = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.parser.ParserError as e:
            raise CliError(f'Error in {file}: {e}')

    if not isinstance(file_inputs, dict):
        raise CliError('Error: Expected input file to contain a dictionary')

    cowait.cli.run(
        ctx.obj,
        task,
        name=name,
        image=image,
        inputs={
            **file_inputs,
            **parse_input_list(input),
        },
        env=parse_input_list(env),
        ports={p: p for p in port},
        routes=parse_input_list(route),
        upstream=upstream,
        build=build,
        detach=detach,
        raw=raw,
        quiet=quiet,
        cpu=cpu,
        cpu_limit=cpu_limit,
        memory=memory,
        memory_limit=memory_limit,
        affinity=affinity,
        cluster_name=cluster,
    )


@click.command(help='run task tests')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-m', '--mount/--no-mount',
              default=True,
              type=bool,
              help='mount working directory')
@click.option('--cpu',
              help='cpu request',
              type=str,
              default=None)
@click.option('--cpu-limit',
              help='cpu limit',
              type=str,
              default=None)
@click.option('--memory', '--mem',
              help='memory request',
              type=str,
              default=None)
@click.option('--memory-limit', '--mem-limit',
              help='memory limit',
              type=str,
              default=None)
@click.pass_context
def test(ctx, cluster: str, mount: bool, cpu: str, cpu_limit: str, memory: str, memory_limit: str):
    cowait.cli.test(
        ctx.obj, cluster_name=cluster, mount=mount,
        cpu=cpu, cpu_limit=cpu_limit, memory=memory, memory_limit=memory_limit
    )


@click.command(help='destroy tasks')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.pass_context
def rm(ctx, cluster: str):
    cowait.cli.destroy(ctx.obj, cluster_name=cluster)


@click.command(help='list tasks')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.pass_context
def ps(ctx, cluster: str):
    cowait.cli.list_tasks(ctx.obj, cluster_name=cluster)


@click.command(help='kill tasks by id')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.argument('task', type=str)
@click.pass_context
def kill(ctx, cluster: str, task: str):
    cowait.cli.kill(ctx.obj, task, cluster_name=cluster)


@click.command(help='deploy cowait agent')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-d', '--detach',
              type=bool, is_flag=True,
              help='run in detached mode',
              default=False)
@click.option('-u', '--upstream',
              type=str, default=None,
              help='custom upstream uri')
@click.pass_context
def agent(ctx, cluster: str, detach: bool, upstream: str):
    cowait.cli.agent(ctx.obj, detach, upstream, cluster_name=cluster)


@click.command(help='stream logs from a running task')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-r', '--raw',
              type=bool, is_flag=True,
              help='raw output',
              default=False)
@click.argument('task', type=str)
@click.pass_context
def logs(ctx, cluster: str, raw: bool, task: str):
    cowait.cli.logs(ctx.obj, task, cluster_name=cluster, raw=raw)



import sys
import json
import getpass
import docker.errors
from cowait.tasks import TaskDefinition
from cowait.engine.errors import TaskCreationError, ProviderError
from cowait.tasks.messages import TASK_INIT, TASK_STATUS, TASK_FAIL, TASK_RETURN, TASK_LOG
from ..config import Config
from ..context import Context
from ..utils import ExitTrap
from ..logger import Logger
from ..task_image import TaskImage
from .build import build as build_cmd
from sty import fg, rs


class TaskFailedError(RuntimeError):
    pass


def run(
    config: Config,
    task: str, *,
    name: str = None,
    image: str = None,
    inputs: dict = {},
    env: dict = {},
    ports: dict = {},
    routes: dict = {},
    volumes: dict = {},
    build: bool = False,
    upstream: str = None,
    detach: bool = False,
    cpu: str = None,
    cpu_limit: str = None,
    memory: str = None,
    memory_limit: str = None,
    raw: bool = False,
    quiet: bool = False,
    affinity: str = None,
    cluster_name: str = None,
):
    logger = RunLogger(raw, quiet)
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)

        # figure out image name
        remote_image = True
        if image is None:
            if build:
                build_cmd(config, quiet=quiet or raw)
            image = context.image
            remote_image = False

        if not isinstance(volumes, dict):
            raise TaskCreationError('Invalid volume configuration')

        # if we are using the image of the current context, automatically mount the working directory
        # todo: add an option to disable this
        if not remote_image:
            volumes['/var/task'] = {
                'bind': {
                    'src': context.root_path,
                    'mode': 'rw',
                    'inherit': 'same-image',
                },
            }

        # default to agent as upstream
        agent = cluster.find_agent()

        # create task definition
        taskdef = TaskDefinition(
            id=name,
            name=task,
            image=image,
            inputs=inputs,
            env={
                **context.extend('environment', env),
                **context.dotenv,
            },
            ports=ports,
            routes=routes,
            parent=None,  # root task
            upstream=context.coalesce('upstream', upstream, agent),
            owner=getpass.getuser(),
            volumes=context.extend('volumes', volumes),
            cpu=context.override('cpu', cpu),
            cpu_limit=context.override('cpu_limit', cpu_limit),
            memory=context.override('memory', memory),
            memory_limit=context.override('memory_limit', memory_limit),
            affinity=context.override('affinity', affinity),
        )

        # print execution info
        logger.print_info(taskdef, cluster)

        # when running in docker, attempt to pull images if they dont exist locally
        if cluster.type == "docker":
            TaskImage.pull(image, tag='latest')

        # submit task to cluster
        task = cluster.spawn(taskdef)

        if detach:
            logger.header('detached')
            return

        def destroy(*args):
            logger.header('interrupt')
            cluster.destroy(task.id)
            sys.exit(1)

        with ExitTrap(destroy):
            # capture & print logs
            logs = cluster.logs(task.id)
            logger.header('task output')
            for msg in logs:
                logger.handle(msg)

        logger.header()

    except docker.errors.NotFound as e:
        logger.print_exception(f'Docker Error: {e.explanation}')
        sys.exit(1)

    except ProviderError as e:
        logger.print_exception(f'Provider Error: {e}')
        sys.exit(1)

    except TaskCreationError as e:
        logger.print_exception(f'Error creating task: {e}')
        sys.exit(1)

    except TaskFailedError:
        sys.exit(1)


class RunLogger(Logger):
    def __init__(self, raw: bool = False, quiet: bool = False, time: bool = True):
        super().__init__(quiet, time)
        self.raw = raw
        self.idlen = 0

    @property
    def newline_indent(self):
        return self.idlen + 4 + super().newline_indent

    def handle(self, msg):
        if 'type' not in msg:
            return
        type = msg['type']

        if self.quiet:
            # only top level return value
            if type == TASK_RETURN and msg['id'] == self.id:
                print(json.dumps(msg['result']))
            return
        elif self.raw:
            print(json.dumps(msg))
        else:
            if type == TASK_INIT:
                self.on_init(**msg)
            elif type == TASK_RETURN:
                self.on_return(**msg)
            elif type == TASK_FAIL:
                self.on_fail(**msg)
            elif type == TASK_STATUS:
                pass
            elif type == TASK_LOG:
                self.on_log(**msg)

    def header(self, title: str = None):
        if self.raw:
            return
        super().header(title)

    def print_info(self, taskdef, cluster):
        self.id = taskdef.id

        self.header('task')
        self.println('   task:      ', self.json(taskdef.id))
        self.println('   cluster:   ', self.json(cluster.type), self.json(cluster.args))
        if taskdef.upstream:
            self.println('   upstream:  ', self.json(taskdef.upstream))
        self.println('   image:     ', self.json(taskdef.image))
        if len(taskdef.inputs) > 0:
            self.println('   inputs:    ', self.json(taskdef.inputs))
        if len(taskdef.volumes) > 0:
            self.println('   volumes:   ', self.json(taskdef.volumes))
        if taskdef.cpu or taskdef.cpu_limit:
            self.println(f'   cpu:        {taskdef.cpu}/{taskdef.cpu_limit}')
        if taskdef.memory or taskdef.memory_limit:
            self.println(f'   memory:     {taskdef.memory}/{taskdef.memory_limit}')

    def print(self, *args):
        if self.raw:
            return
        super().print(*args)

    def print_id(self, id, ts=None, short=True, pad=True):
        color = fg(hash(id) % 214 + 17)
        if short and '-' in id:
            id = id[:id.rfind('-')]
            self.idlen = max(self.idlen, len(id))
        self.print(color + id.ljust(self.idlen if pad else 0) + rs.all)

    def on_init(self, task: dict, version: str, ts: str = None, **msg):
        self.print_time(ts)
        self.print_id(task['id'])
        self.print(
            f' {fg.yellow}*{rs.all} started with',
            self.json(task['inputs'], indent=2),
        )
        if task['parent'] is not None:
            self.print(' by [')
            self.print_id(task['parent'], pad=False)
            self.println(']')
        else:
            self.println()

    def on_status(self, id: str, status: str, ts: str = None, **msg):
        self.print_time(ts)
        self.print_id(id)
        self.println(f'{fg.yellow} ~ {status}{rs.all}')

    def on_fail(self, id: str, error: str, ts: str = None, **msg):
        self.print_time(ts)
        self.print_id(id)
        self.println(f'{fg.red} ! {rs.all}ERROR: {error}')
        if id == self.id:
            raise TaskFailedError(error)

    def on_return(self, id: str, result: any, ts: str = None, **msg):
        self.print_time(ts)
        self.print_id(id)
        self.println(f'{fg.green} ={rs.all} returned', self.json(result, indent=2))

    def on_log(self, id: str, file: str, data: str, ts: str = None, **msg):
        self.print_time(ts)
        self.print_id(id)
        self.println('  ', data)



import random
from .router import Router

STATIC_PORTS = {
    'cowait.tasks.agent': 1339,
    'cowait.notebook': 1889,
}


class LocalPortRouter(Router):
    def __init__(self, cluster):
        super().__init__(cluster)
        cluster.on('prepare', self.on_prepare)

    def on_prepare(self, taskdef):
        for path, port in taskdef.routes.items():
            if path != '/':
                print(f'!! warning: http subdirectory path {path} '
                      'is not routable locally')
                continue

            host_port = STATIC_PORTS.get(taskdef.name, random.randint(60000, 65000))

            url = f'http://localhost:{host_port}/'

            # assign route url
            taskdef.routes[path] = {
                'port': port,
                'path': path,
                'url': url,
            }

            # open host port
            taskdef.ports[port] = host_port

            print('~~ http route', path, '->', url)

        return taskdef



from .type import Type
from .simple import Any
from .mapping import is_cowait_type, convert_type, TypeAlias


@TypeAlias(list)
@TypeAlias(tuple)
class List(Type):
    """ List type """

    def __init__(self, item: Type = Any()):
        if not is_cowait_type(item):
            raise ValueError('Element type is not a cowait Type')

        self.item = convert_type(item)

    def validate(self, value: list, name: str) -> None:
        if not isinstance(value, list):
            raise ValueError(f'{name} is not a list')

        for i, item in enumerate(value):
            self.item.validate(item, f'{name}[{i}]')

    def serialize(self, value: list) -> list:
        return [self.item.serialize(item) for item in value]

    def deserialize(self, value: list) -> list:
        return [self.item.deserialize(item) for item in value]

    def describe(self):
        item_desc = self.item.describe()
        if item_desc is not None:
            return [item_desc]
        return []



import asyncio
from cowait.types import deserialize
from concurrent.futures import Future
from cowait.tasks.components.rpc import RpcError, \
    RPC_CALL, RPC_ERROR, RPC_RESULT


class RpcCall(Future):
    def __init__(self, method, args, nonce):
        super().__init__()
        self.method = method
        self.args = args
        self.nonce = nonce


class RpcClient(object):
    def __init__(self, ws):
        self.ws = ws
        self.nonce = 0
        self.calls = {}

    async def call(self, method, args):
        nonce = self.nonce
        self.nonce += 1

        call = RpcCall(method, args, nonce)

        # track pending rpc calls
        self.calls[nonce] = call

        await self.ws.send_json({
            'type':   RPC_CALL,
            'method': method,
            'args':   args,
            'nonce':  nonce,
        })

        return await asyncio.wrap_future(call)

    def cancel_all(self):
        for future in self.calls.values():
            if not future.done():
                future.set_exception(RpcError('Cancelled'))

    def intercept_event(self, type, **msg):
        if type == RPC_RESULT:
            self._rpc_result(**msg)
            return True

        if type == RPC_ERROR:
            self._rpc_error(**msg)
            return True

        return False

    def _rpc_result(self, nonce, result, result_type, **msg):
        result = deserialize(result, result_type)
        future = self.calls[nonce]
        if not future.done():
            future.set_result(result)
        del self.calls[nonce]

    def _rpc_error(self, nonce, error, **msg):
        future = self.calls[nonce]
        if not future.done():
            future.set_exception(RpcError(error))
        del self.calls[nonce]



import os
from cowait.tasks import TaskDefinition
from cowait.engine.const import ENV_TASK_DEFINITION, MAX_ENV_LENGTH
from cowait.engine.utils import env_unpack, base_environment
from cowait.engine.errors import ProviderError


def create_ports(taskdef):
    return taskdef.ports


def create_env(cluster, taskdef):
    env = base_environment(cluster, taskdef)
    
    # check total length of environment data
    length = 0
    for key, value in env.items():
        if isinstance(value, dict):
            # complex env settings are not supported by DockerProvider
            # try to inherit the setting from the host
            if key in os.environ:
                value = os.environ[key]
            else:
                source = value.get('source', '<unset>')
                print(f'Warning: unset environment variable {key} with source "{source}"')
                value = ''

        length += len(str(key)) + len(str(value))
        env[key] = str(value)

    if length > MAX_ENV_LENGTH:
        raise ProviderError(f'Task environment too long. Was {length}, max: {MAX_ENV_LENGTH}')

    return env


def extract_container_taskdef(container) -> TaskDefinition:
    for env in container.attrs['Config']['Env']:
        if ENV_TASK_DEFINITION == env[0:len(ENV_TASK_DEFINITION)]:
            data = env[len(ENV_TASK_DEFINITION)+1:]
            return TaskDefinition(**env_unpack(data))
    raise Exception('Unable to unpack container task definition')



from pyspark.conf import SparkConf
from cowait.network import get_local_ip
from cowait.types import Type, TypeAlias


@TypeAlias(SparkConf)
class SparkConfType(Type):
    name: str = 'SparkConf'

    def validate(self, conf: SparkConf, name: str) -> None:
        if not isinstance(conf, dict):
            raise ValueError(f'{name} is not a SparkConf')

    def serialize(self, conf: SparkConf):
        # serialize spark configuration
        return {key: value for key, value in conf.getAll()}

    def deserialize(self, config: dict):
        conf = SparkConf()
        for option, value in config.items():
            conf.set(option, value)

        # set driver host to our local ip
        conf.set('spark.driver.host', get_local_ip())

        return conf



import os
import cowait
import asyncio
from ipykernel.kernelapp import IPKernelApp
from ipykernel.ipkernel import IPythonKernel
from cowait.tasks import Task, TaskDefinition
from cowait.tasks.notebook import NotebookRunner
from cowait.network import get_local_url
from cowait.worker import env_get_cluster_provider, env_get_task_definition
from .node import NotebookNode

import cowait.notebook.html_repr  # noqa: F401

ENV_KERNEL_TOKEN = 'COWAIT_KERNEL_TOKEN'


class KernelTask(Task):
    pass


class CowaitKernel(IPythonKernel):
    def start(self):
        super().start()
        asyncio.get_event_loop().create_task(self.init_node())

    async def init_node(self):
        cluster = env_get_cluster_provider()
        parent = env_get_task_definition()
        token = os.getenv(ENV_KERNEL_TOKEN)

        taskdef = TaskDefinition(
            name='kernel',
            image=parent.image,
            parent=parent.id,
            volumes=parent.volumes,
            env=parent.env,
            upstream=get_local_url(),
            meta={
                **parent.meta,
                'virtual': True,
            },
        )

        # set up notebook node
        self.node = NotebookNode(taskdef)
        await self.node.start(token)

        # instantiate kernel task
        self.task = KernelTask(node=self.node, cluster=cluster, taskdef=taskdef)

        # write globals
        self.shell.push({
            'cowait': cowait,
            'kernel': self.task,
            'tasks': self.task.subtasks,
            'NotebookRunner': NotebookRunner,
        })

    def do_shutdown(self, restart):
        asyncio.get_event_loop().create_task(self.shutdown_node())
        super().do_shutdown(restart)

    async def shutdown_node(self):
        await self.node.stop()
        self.node = None
        self.task = None


if __name__ == '__main__':
    app = IPKernelApp.instance(kernel_class=CowaitKernel)
    app.initialize()
    app.start()



import sys
from io import StringIO


class StreamCapture(object):
    def __init__(
        self,
        stream,
        silence: bool = False,
        callback: callable = None,
    ):
        self.stream = stream
        self.silence = silence
        self.capture = StringIO()
        self.callback = callback

    def isatty(self):
        return False

    def write(self, data):
        self.capture.write(data)

        if not self.silence:
            self.stream.write(data)

        if '\n' in data:
            self.flush(auto=True)

    def flush(self, auto: bool = False):
        if not self.silence:
            self.stream.flush()

        if callable(self.callback):
            self.callback(self.getvalue())

        self.capture = StringIO()

    def getvalue(self):
        data = self.capture.getvalue()
        return data


class StreamCapturing(object):
    def __init__(
        self,
        on_stdout: callable = None,
        on_stderr: callable = None,
        silence=False,
    ):
        self.on_stdout = on_stdout
        self.on_stderr = on_stderr
        self.silence = silence

    def __enter__(self):
        self.stdout = StreamCapture(
            stream=sys.stdout,
            callback=self.on_stdout,
            silence=self.silence,
        )
        self.stderr = StreamCapture(
            stream=sys.stderr,
            callback=self.on_stderr,
            silence=self.silence,
        )
        # capture outputs
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        return self

    def __exit__(self, *args):
        # reset outputs
        sys.stdout = self.stdout.stream
        sys.stderr = self.stderr.stream

    def __del__(self):
        del self.stdout
        del self.stderr
